/****
 *	Copyright (c) 2007, Olivier Marin, Laboratoire d'Informatique de Paris 6
 *	All rights reserved.
 ****/

/* 
	This file is part of DARX.

	DARX is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	DARX is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with DARX. If not, see <http://www.gnu.org/licenses/>.
 */




package darx;

import java.rmi.RemoteException;
import java.util.*;
import java.io.*;


/**
 * This is the thread launched every time a synchronous message delivery
 * is attempted. Its goal is to provide a single reply for a request forwarded
 * to all the active replicas within the replication group.
 */
class SyncMessageDeliver extends Thread {
    private TaskShellHandle shell;
    private DarxMessage msg;
    private Serializable reply;
    private RemoteException e;

    public SyncMessageDeliver(TaskShellHandle shell, DarxMessage msg) {
	this.shell = shell;
	this.msg = msg;
	e = null;
    }

    public void run() {
	try {
	    reply = shell.processDeliverSyncMessage(msg);
	}
	catch(RemoteException e) {
	    reply = null;
	    this.e = e;
	}
    }

    public Serializable getReply() { return reply; }

    public RemoteException getException() { return e; }
}



/**
 * This is the replication group manager specialisation which implements the
 * active replication strategy.
 *
 * @author Jacob Zimmermann
 * @author Olivier Marin
 *
 * @version %I%, %G%
 */
public class ActiveReplicationStrategy extends ReplicationStrategy 
    implements Serializable {


    public ActiveReplicationStrategy() {
	super();
	type = ReplicationStrategy.ACTIVE_STRATEGY;
    }


    public ActiveReplicationStrategy(ActiveReplicationStrategy newstr) {
	super();
	info = newstr.getReplicantInfo();
	replicants = newstr.getReplicants();
	rep_number = newstr.getReplicationNumber();
    }


    public int compareTo(ReplicationStrategy other_strat) {
	int other_strat_type = other_strat.getReplicationType();
	return (type - other_strat_type);
    }


    public void suspend() {
	for (Enumeration en = replicants.elements(); en.hasMoreElements();)
	    try {
		((TaskShellHandle) en.nextElement()).suspend();
	    }
	    catch(RemoteException e) {
		handleUnreachableReplicant(info, e);
	    }
    }


    public void resume() {
	for (Enumeration en = replicants.elements(); en.hasMoreElements();)
	    try {
		((TaskShellHandle) en.nextElement()).resume();
	    }
	    catch(RemoteException e) {
		handleUnreachableReplicant(info, e);
	    }
    }


    public void deliverAsyncMessage(TaskShell shell, DarxMessage msg) {
	int k;
	try {
	    shell.processDeliverAsyncMessage(msg);
	}
	catch(Exception e) {
	    System.out.println("Couldn't deliver asynchronous message #"
			       +msg.getSerial()
			       +" from "
			       +msg.getSenderName());
	}
	/*if(shell.acceptMsg(msg)) {
	  Thread deliver=new MessageDeliverThread(msg, shell.task);
	  deliver.start();
	  }*/
	for (Enumeration en = replicants.elements(); en.hasMoreElements();)
	    try {
		((TaskShellHandle)en.nextElement())
		    .processDeliverAsyncMessage(msg);
	    }
	    catch(RemoteException e) {
		handleUnreachableReplicant(info, e);
	    }
    }
    

    public Serializable deliverSyncMessage(TaskShell shell, DarxMessage msg) {
	SyncMessageDeliver[] proc = new SyncMessageDeliver[replicants.size()+1];
	proc[0] = new SyncMessageDeliver(shell, msg);
	proc[0].start();
	int k = 0;
	TaskShellHandle handle;
	for (Enumeration en=replicants.elements(); en.hasMoreElements();) {
	    handle = (TaskShellHandle)en.nextElement();
	    proc[k+1] = new SyncMessageDeliver(handle, msg);
	    proc[k+1].start();
	    k++;
	}
	for(k = 0;  k < proc.length; k++) {
	    try {
		proc[k].join();
		if(proc[k].getException() != null)
		    handleUnreachableReplicant(info, proc[k].getException());
	    }
	    catch(InterruptedException e) {}
	}
	return proc[0].getReply();
    }


}
