/****
 *	Copyright (c) 2007, Olivier Marin, Laboratoire d'Informatique de Paris 6
 *	All rights reserved.
 ****/

/* 
	This file is part of DARX.

	DARX is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	DARX is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with DARX. If not, see <http://www.gnu.org/licenses/>.
 */





package darx;

import java.rmi.*;
import java.rmi.server.*;
import java.net.*;
import java.rmi.registry.*;


/**
 * This is the server implementation which, as main objective, provides
 * the means for running DARX tasks on the local host.
 * 
 * 
 * @author Jacob Zimmermann
 * @author Olivier Marin
 * 
 * @version %I%, %G%
 */ 
public class Darx extends UnicastRemoteObject implements DarxServer {

    /**
     * The stub providing access to the name server functionalities.
     */
    static protected NameServer nserver = null;

    /**
     * The URL of the name server.
     */
    static protected String ns_url = NameServer.DEFAULT_URL;

    /**
     * The port number of the name server.
     */
    static protected int ns_port_nb = NameServer.DEFAULT_PORT_NB;

    /**
     * The RMIregistry serving this specific location (<host>:<port_nb>).
     */
    static protected Registry local_registry;

    /**
     * The local server's URL
     */
    static protected String myURL = "";

    /**
     * The local server's name
     */
    static protected String myName;

    /**
     * The local server's port number
     */
    static protected int myPortNb = 6789;

    /**
     * The generic service name
     */
    static public final String SERVICE_NAME = "DarxServer";




    //------------------------------------//
    //---- SERVER CREATION & ROUTINES ----//
    //------------------------------------//




    /**
     * Constructs a new DARX server instance.
     */
    public Darx() throws RemoteException {
    }


    /*
     * Creation of the server and its running environment.
     */
    static public void main(String[] params) {
	try {
	    analyseCommandLine(params);
	    //Retrieve reference to the naming service
	    Registry nserver_reg = 
		LocateRegistry.getRegistry(ns_url, ns_port_nb);
	    nserver = (NameServer) 
		nserver_reg.lookup(ns_url+":"+ns_port_nb+"/NameServer");
	    System.out.println("NameServer registry found at "
			       + ns_url + ":" + ns_port_nb);
	    //Retrieve URL
	    myURL = InetAddress.getLocalHost().getHostName();
	}
	catch(Exception e) {
	    System.out.println("DARX server> Name Server URL : " + ns_url);
	    System.out.println("Error : cannot connect to nameserver : "+e);
	    e.printStackTrace();
	    System.exit(0);
	}
	try {
	    Darx darx = new Darx();
	    //Create new RMI registry on the location
	    local_registry = LocateRegistry.createRegistry(myPortNb);
	    //Register server
	    myName = myURL + ":" + myPortNb + "/" + SERVICE_NAME;
	    local_registry.rebind(myName, darx);
	    System.out.println("Darx server ready at "+ myURL + ":" + myPortNb);
	}
	catch(Exception e) {
	    System.out.println("Error during initialisation : "+e);
	    System.exit(0);
	}
    }


    /**
     * Analyses the command line in order to fill in specific runtime info: the 
     * URL and port number of the name server if the default value is not used,
     * as well as the port number of the DARX server if the default value (6789)
     * is not used.
     *
     * @param parameters  the String array containing the parameters, can be
     *                    void if default values are used and debug mode is off
     */
    static void analyseCommandLine(String[] parameters) {
	int l = parameters.length;
	int i = 0;
	while (i < l) {
	    if (parameters[i].compareTo("-ns") == 0) {
		i++;
		ns_url = parameters[i];
		NameServerImpl.URL = ns_url;
		i++;
		ns_port_nb = (new Integer(parameters[i])).intValue();
		NameServerImpl.PORT_NB = ns_port_nb;
		i++;
	    } else if (parameters[i].compareTo("-p") == 0) {
		i++;
		myPortNb = (new Integer(parameters[i])).intValue();
		i++;
	    } else {
		System.out.println("Incorrect use of DARX server startup command:"
				   + "\n'./startdarx -ns <ns_url ns_port_number>] "
				   +"[-p <port_number>]'");
		System.exit(0);
	    }
	}
    }


    //--------------------------------------//
    //---- TASK RMI REGISTRY MANAGEMENT ----//
    //--------------------------------------//

    static private void rebindTask(ReplicantInfo rep_info, TaskShell shell) 
	throws RemoteException {
	Registry r = LocateRegistry.getRegistry(rep_info.getPortNb());
	r.rebind(rep_info.textifyDarxName(), shell);
    }


    static private void unbindTask(ReplicantInfo rep_info) 
	throws RemoteException {
	Registry r = LocateRegistry.getRegistry(rep_info.getPortNb());
	try {
	    System.out.println("Unbinding " + rep_info.textifyDarxName());
	    r.unbind(rep_info.textifyDarxName());
	} catch (NotBoundException e) {
	    System.out.println("Task " + rep_info.textifyDarxName()
			       + " isn't bound by the naming service");
	}
    }



    //-----------------------------------//
    //---- TASK EXECUTION MANAGEMENT ----//
    //-----------------------------------//



    /**
     * Starts the execution of a new replication group leader on the current
     * server.
     *
     * @param task  the <code>DarxTask</code> to be started on this server.
     * @return  the handle for the started task; to be used for user purposes,
     *          such as (un)replication or policy modification.
     */
    public RemoteTask startTask(DarxTask task) 
	throws RemoteException {
	DarxHandle handle=null;
	String generic_task_name = task.getTaskName();
	System.out.println("Starting task " + generic_task_name);
	//Generate the info for the new replicant
	ReplicantInfo new_rep_info = 
	    new ReplicantInfo(myURL, myPortNb, generic_task_name, 0);
	//Create the shell for the new replicant 
	//& set it as the leader of its group
	TaskShell shell = new TaskShell(task, new_rep_info);
	shell.setAsGroupLeader();
	try {
	    //Register new task locally, then on name server
	    rebindTask(new_rep_info, shell);
	    nserver.register(new_rep_info);
	    //Start task execution
	    task.start();
 	    //Check task registration
 	    String handle_name = new_rep_info.textifyDarxName();
 	    handle = (DarxHandle) local_registry.lookup(handle_name);
 	    handle = shell;	    
	} 
	catch(NotBoundException e) {
	    System.out.println("Error : " + new_rep_info.textifyDarxName() 
		+ " is not bound ?!?!");
	}
	System.out.println(generic_task_name+" started...\n");
	RemoteTask remote = new RemoteTask(new_rep_info, handle);
	task.setHandle(remote);
	return remote;
    }


    /**
     * Creates a task replica on the current server.
     * This method is called by the <code>TaskShell</code> of the  group
     * leader when it has to be replicated.
     *
     * @param info  the <code>ReplicantInfo</code> of the new replica.
     * @param task  the <code>DarxTask</code> to be replicated on this server.
     * @return the remote reference to the newly created replica
     */ 
    public TaskShellHandle createReplicant(ReplicantInfo info, DarxTask task)
	throws RemoteException {
	String rep_name = info.textifyDarxName();
	System.out.println("Creating replicant '" + rep_name + "'");
	TaskShell shell = new TaskShell(task, info);
	//Register the new replica on the naming service
	nserver.register(info);
	System.out.println(info.textifyDarxPath() + " created");
	//Register the new replica on the local RMI registry
	rebindTask(info, shell);
	System.out.println(rep_name + " registered locally\n");
	//Retrieve the remote reference to the new replica
	TaskShellHandle rep_handle = null;
	try {
	    rep_handle = (TaskShellHandle)local_registry.lookup(rep_name);
	} catch(NotBoundException e) {
	    System.out.println("Can't find remote reference to " 
			       + rep_name + " on the local RMI registry");
	    e.printStackTrace();
	}
	return rep_handle;
    }


    /**
     * Destroys the local replicant corresponding to the specified task.
     * This includes unreferencing the doomed replicant on the local RMI server
     * and on the name server.
     *
     * @param task_name  the name of the task to which the replicant belongs
     */
    public void killReplicant(String task_name) throws RemoteException {
	System.out.println("Treating deletion request for task: " + task_name);
	ReplicantInfo rep_info = null;
	try {
	    //Remove the doomed replicant info from the NameServer
	    rep_info = nserver.unregister(task_name, myURL, myPortNb);
	    //Retrieve the local reference to the local replicant
	    String hn = rep_info.textifyDarxName();
	    TaskShellHandle shell = (TaskShellHandle)local_registry.lookup(hn);
	    //Remove the reference from the local RMIRegistry
	    unbindTask(rep_info);
	    //Terminate the replicant execution
	    shell.terminateTask();
	}
	catch(NotBoundException e2) {
	    System.out.println("No known replicant for task: " 
			       + task_name + " on this server");
	}
	if (rep_info != null)
	    System.out.println(rep_info.textifyDarxName() 
			       + " successfully deleted\n");
    }


    //-----------------------//
    //---- TASK SERVICES ----//
    //-----------------------//


    /**
     * Generates a remote reference to the current leader of the task
     * which corresponds to the given generic name.
     * If the current leader cannot be contacted, it is assumed that the
     * host has failed, and therefore a new leader election is launched 
     * through the <code>NameServer</code> services.
     **/
    public RemoteTask findTask(String task_name) throws RemoteException {
	try {
	    ReplicantInfo task_info = nserver.getLeaderOfTask(task_name);
	    
	    boolean next = false;
	    TaskShellHandle ts_handle = null;
	    //get the DARX server supporting the current task leader
	    try {
		ts_handle = task_info.getTaskShellHandle();
	    } catch(Exception e) { next = true; }
	    //This loop is entered if a non-DARX exception was caught
	    while(next) {
		try {
		    task_info = nserver.selectAnotherLeader(task_info);
		    ts_handle = task_info.getTaskShellHandle();
		    ts_handle.setAsGroupLeader();
		    next = false;
		}
		catch(DarxException e) {
		    throw e;
		}
		catch(Exception e) { next = true; }
	    }
	    return new RemoteTask(task_info, ts_handle);
	}
	catch(DarxException e) {
	    throw e;
	}
	catch(RemoteException e) {
	    System.out.println("Internal error : "+e);
	}
	return null;
    }




}
