/****
 *	Copyright (c) 2007, Olivier Marin, Laboratoire d'Informatique de Paris 6
 *	All rights reserved.
 ****/

/* 
	This file is part of DARX.

	DARX is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	DARX is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with DARX. If not, see <http://www.gnu.org/licenses/>.
 */






package darx;

import java.io.*;
import java.rmi.RemoteException;
import java.util.*;


/**
 * This is the replication group manager specialisation which implements the
 * passive replication strategy.
 * 
 * @author Jacob Zimmermann
 * @author Olivier Marin
 * 
 * @version %I%, %G%
 */ 
public class PassiveReplicationStrategy 
    extends ReplicationStrategy 
    implements Runnable, Serializable {

    /**
     * The update delay, that is the time elapsed between backups.
     */
    private int updateDelay;

    /**
     * The thread used to update the replicas concurrently to the
     * leader's execution.
     */
    private transient Thread updater;

    /**
     * The DARX internal handle for the replica which owns this instance.
     */
    private transient TaskShellHandle owner_handle = null;

    /**
     * The state of the thread (running, suspended, or dead).
     */
    private int updater_state;

    /**
     * The state of the thread as: running.
     */
    private final int RUNNING = 0;

    /**
     * The state of the thread as: suspended.
     */
    private final int SUSPENDED = 1;

    /**
     * The state of the thread as: dead.
     */
    private final int DEAD = 2;



    //CONSTRUCTION


    /**
     * Constructs a new instance of the passive replication strategy.
     *
     * @param ud  the delay between every backup.
     */
    public PassiveReplicationStrategy(int ud) {
	super();
	type = ReplicationStrategy.PASSIVE_STRATEGY;
	updateDelay = ud;
    }


    public int compareTo(ReplicationStrategy other_strat) {
	int other_strat_type = other_strat.getReplicationType();
	int result = (type - other_strat_type);
	if (result == 0)
	    result = ((PassiveReplicationStrategy)other_strat).getUpdateDelay()
		- updateDelay;
	return result;
    }



    //SELECTION & MODIFICATION

    /**
     * Returns the update delay.
     *
     * @return  the update delay.
     */
    public int getUpdateDelay() {
	return updateDelay;
    }

    /**
     * Sets the update delay.
     *
     * @param ud  the new update delay.
     */
    public void setUpdateDelay(int ud) {
	updateDelay = ud;
    }


    //BACKUP HANDLING


    /**
     * Updates the replication group.
     * Sends a fresh copy of the leading <code>DarxTask</code> to replace
     * the one held in the <code>TaskShell</code> of the backups.
     */
    protected void update() {
	System.out.println("Updating replicas...");
	int k;
	if (owner_handle == null)
	    owner_handle = getRemoteShellHandle(info);
	DarxTask task = null;
	TaskShellHandle handle = null;
	try {
	  //Get the local task to be exported to the backups
	  //I expect this call to provide a deep copy of the original DarxTask 
	  //through serialization; please contact me if I'm wrong
	    //NEED TO SUSPEND DARX_TASK_ENGINE HERE
	  task = (DarxTask)owner_handle.getTask();
	    //NEED TO RESUME DARX_TASK_ENGINE HERE
	  //To satisfy both needs, necessity calls for a getConsistentTaskImage
	  //that will remotely suspend, deep-copy and then resume
	}
	catch(RemoteException re) {
	    System.out.println("Error while updating replicas, couldn't access"
			       +" the replicated task's state: "+ re);
	}
	for (Enumeration e = replicants.elements() ; e.hasMoreElements() ;) {
	    try {
		handle = (TaskShellHandle) e.nextElement();
		handle.setTask(task);
		handle.demoteFromGroupLeader();
	    }
	    catch(RemoteException re) {
		handleUnreachableReplicant(info, re);
	    }
	}
	System.out.println(".............. update done");
    }


    //EXECUTION HANDLING


    /**
     * Starts the backup process.
     * This method will only be called if the owner of the strategy is a leader.
     *
     * @param ud  the update delay.
     * @see #resume()
     */
    public void activateUpdating(int ud) {
	updateDelay = ud;
	if (updater == null) {
	    updater = new Thread(this);
	    updater.start();
	}
	updater_state = RUNNING;
    }


    public void suspend() {
	System.out.println("<PassiveReplicationStrategy> Suspending strategy");
// 	if (role == ReplicationStrategy.LEADER && updater != null)
	    update();
	updater_state = SUSPENDED;
    }


    public void resume() {
// 	if (role == ReplicationStrategy.LEADER) {
	    System.out.println("<PassiveReplicationStrategy> Resuming strategy");
	    activateUpdating(updateDelay);
// 	}
    }


    /**
     * Executes the backup mechanism.
     */
    public void run() {
	while (updater_state != DEAD) {
	    try {
		updater.sleep(updateDelay);
	    }
	    catch(InterruptedException e) {}
	    if (updater_state == RUNNING) 
// 		&& (role == ReplicationStrategy.LEADER))
		update();
	}
    }


    public void stop() {
	updater_state = DEAD;
	updater = null;
    }	


    public void terminate() throws RemoteException {
	stop();
	super.terminate();
    }


    //COMMUNICATION HANDLING


    public void deliverAsyncMessage(TaskShell shell, DarxMessage msg) {
	try {
	    shell.processDeliverAsyncMessage(msg);
	}
	catch(RemoteException e) {}	
    }


    public Serializable deliverSyncMessage(TaskShell shell, DarxMessage msg) {
	Serializable reply = null;
	try {
	    reply = shell.processDeliverSyncMessage(msg);
	}
	catch(RemoteException e) {}
	return reply;
    }


}
