/****
 *	Copyright (c) 2007, Olivier Marin, Laboratoire d'Informatique de Paris 6
 *	All rights reserved.
 ****/

/* 
	This file is part of DARX.

	DARX is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	DARX is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with DARX. If not, see <http://www.gnu.org/licenses/>.
 */





package darx;

import java.rmi.*;
import java.rmi.registry.*;
import java.io.Serializable;
import java.net.MalformedURLException;


/**
 * This is the local proxy that allows user application tasks to handle
 * other tasks.
 * This class defines the core services Darx provides to user applications
 * for handling the different tasks (or agents) of the distributed system.
 * It includes replication management and message sending. At length, the 
 * replication management scheme should become transparent to the user,
 * and therefore might disappear from the current services provided herein.
 * 
 * @author Jacob Zimmermann
 * @author Olivier Marin
 * 
 * @version %I%, %G%
 */ 
public class RemoteTask implements Serializable {


    /**
     * The handle for the leader of the group which represents the task
     */
    private DarxHandle handle;

    /**
     * The information concerning the leader of the task
     */
    private ReplicantInfo leader_info;

    /**
     * the generic name of the task
     */
    private String task_name;



    /**
     * Constructs a user-sufficient definition of a remote task.
     * This is the unique class constructor. It is not supposed to be called
     * from user applications, however; instances of this class are obtained
     * when starting the corresponding DarxTask.
     *
     * @param new_leader_info  the information concerning the leader of the
     *                         task which this instance references.
     * @param handle           the interface to access the leader in charge of
     *                         this task.
     * @see DarxHandle
     * @see TaskShell
     */
    RemoteTask(ReplicantInfo new_leader_info, DarxHandle handle) {
	leader_info = new_leader_info;
	task_name = leader_info.getTaskName();
	this.handle = handle;
    }


    /**
     * Constructs a user-sufficient definition of a remote task.
     * This is the unique class constructor. It is not supposed to be called
     * from user applications, however; instances of this class are obtained
     * when starting the corresponding DarxTask.
     *
     * @param new_leader_info  the information concerning the leader of the
     *                         task which this instance references.
     * @see DarxHandle
     * @see TaskShell
     */
    RemoteTask(ReplicantInfo new_leader_info) throws RemoteException {
	leader_info = new_leader_info;
	task_name = leader_info.getTaskName();
	handle = leader_info.getTaskShellHandle();
    }


    /**
     * Sets the information concerning the leader of the remote task.
     *
     * @param new_leader_info  the information concerning the current leader.
     */
    void setLeaderInfo(ReplicantInfo new_leader_info) throws RemoteException {
	if (leader_info != new_leader_info) {
	    leader_info = new_leader_info;
	    task_name = leader_info.getTaskName();
	    handle = leader_info.getTaskShellHandle();
	}
    }


    /**
     * Sends a message asynchronously to the remote task, that is to the 
     * leader of the corresponding replication group.
     * If the first emission fails, the thread is interrupted for a while
     * and a second emission attempt is initiated. On a second failure, a
     * new leader selection is asked from the <code>NameServer</code> and
     * the whole emission procedure is restarted.
     *
     * @param msg  the contents of the emission.
     *
     * @see DarxHandle#deliverAsyncMessage(DarxMessage msg)
     */
    void sendAsyncMessage(DarxMessage msg) {
	try {
	    handle.deliverAsyncMessage(msg);
	} catch(RemoteException e) {
	    System.out.println("Asynchronous message not delivered");
	    try {
		System.out.println("Trying emission once more...");
		//Take a nap
		java.lang.Thread.sleep(4000);
		//Try emitting again
		handle.deliverAsyncMessage(msg);
	    } catch(InterruptedException e1) {
		System.out.println("Can't go to sleep before reemission!");
		e1.printStackTrace();
	    } catch(RemoteException e2) {
		System.out.println("Asynchronous message not delivered");
		e2.printStackTrace();
		System.out.println("Searching other replicant");
		try {
		    findNewLeader();
		} catch(RemoteException e3) {
		    System.out.println("Can't find new leader!");
		    e3.printStackTrace();
		    System.exit(1);
		}
		//New leader was found: restart message emission
		sendAsyncMessage(msg);
	    }
	}
    }


    /**
     * Sends a message synchronously to the remote task, that is to the 
     * leader of the corresponding replication group.
     *
     * @param msg  the contents of the emission.
     *
     * @see DarxHandle#deliverSyncMessage(DarxMessage msg)
     */
    Object sendSyncMessage(DarxMessage msg) {
	Object reply = null;
	try {
	    reply = handle.deliverSyncMessage(msg);
	} catch(RemoteException e) {
	    System.out.println("Synchronous message cannot be delivered : "+e);
	    System.out.println("Trying emission once more...");
	    try {
		//Take a nap
		java.lang.Thread.sleep(4000);
		//Try emitting again
		reply = handle.deliverSyncMessage(msg);
	    } catch(InterruptedException e1) {
		System.out.println("Can't go to sleep before reemission!");
		e1.printStackTrace();
	    } catch(RemoteException e2) {
		System.out.println("Asynchronous message not delivered");
		e2.printStackTrace();
		System.out.println("Searching other replicant");
		try {
		    findNewLeader();
		} catch(RemoteException e3) {
		    System.out.println("Can't find new leader!");
		    e3.printStackTrace();
		    System.exit(1);
		}
		//New leader was found: restart message emission
		reply = sendSyncMessage(msg);
	    }
	}
    return reply;
}


    /**
     * Starts the mechanism to create a new replica of the remote task at a
     * specified URL.
     * The default port (6789) is used and the default (active) replication
     * strategy is applied.
     * 
     * @param url  the URL of the remote DARX server where the new replica is
     *             created
     * @see DarxHandle
     * @see Darx#createReplicant
     * @see TaskShell#replicateTo
     */
    public void replicateTo(String url) throws RemoteException {
	handle.replicateTo(url, 6789, new ActiveReplicationStrategy()); 
    }


    /**
     * Starts the mechanism to create a new replica of the remote task on a
     * specified server.
     * The default (active) replication strategy is applied.
     * This method enables to specify the server's port number; it is mandatory
     * to use it instead of its homonym when there are several DARX servers
     * concurrently running on the same host.
     * 
     * @param url      the URL of the remote DARX server where the new replica
     *                 is created
     * @param port_nb  the port number of the remote DARX server where the new
     *                 replica is created
     * @see DarxHandle
     * @see Darx#createReplicant
     * @see TaskShell#replicateTo
     */
    public void replicateTo(String url, int port_nb) throws RemoteException {
	handle.replicateTo(url, port_nb, new ActiveReplicationStrategy()); 
    }


    /**
     * Starts the mechanism to create a new replica of the remote task on a
     * specified server following a given replication strategy.
     * This method enables to specify the server's port number; it is mandatory
     * to use it instead of its homonym when there are several DARX servers
     * concurrently running on the same host.
     * 
     * @param url      the URL of the remote DARX server where the new replica
     *                 is created
     * @param port_nb  the port number of the remote DARX server where the new
     *                 replica is created
     * @param rs       the replication strategy applied to the new replica
     * @see DarxHandle
     * @see Darx#createReplicant
     * @see TaskShell#replicateTo
     */
    public void replicateTo(String url, int port_nb, ReplicationStrategy rs) 
	throws RemoteException, IllegalReplicationException {
	handle.replicateTo(url, port_nb, rs); 
    }


    /**
     * Starts the mechanism to destroy a specific replica of the remote task on
     * a given server.
     * 
     * @param url  the URL of the remote DARX server hosting the doomed replica
     * @exception RemoteException
     * @see DarxHandle
     */
    public void killReplicantAt(String url) 
	throws RemoteException, UnknownReplicantException {
	handle.killReplicantAt(url, 6789); 
    }


    /**
     * Starts the mechanism to destroy a specific replica of the remote task on
     * a given server.
     * This method enables to specify the server's port number; it is mandatory
     * to use it instead of its homonym when there are several DARX servers
     * concurrently running on the same host.
     * 
     * @param url      the URL of the remote DARX server supporting 
     *                 the doomed replica
     * @param port_nb  the port number of the same remote DARX server
     * @exception RemoteException
     * @see DarxHandle
     */
    public void killReplicantAt(String url, int port_nb) 
	throws RemoteException, UnknownReplicantException {
	handle.killReplicantAt(url, port_nb); 
    }


    /**
     * Terminates the execution of this replication group.
     */
    public void killTask() throws RemoteException {
	handle.killTask();
    }


    /**
     * Modifies the current replication policy of the group, by applying
     * a new strategy to a given replica.
     *
     * @param url       the URL of the replica to which the new strategy applies
     * @param port_nb   the port number of the replica to which the new strategy
     *                  applies
     * @param strategy  the new strategy to be applied
     * @see DarxHandle
     * @see ReplicationStrategy
     * @see TaskShell#switchReplicationStrategy
     */
    public void switchReplicationStrategy(String url,
					  int port_nb,
					  ReplicationStrategy strategy) 
	throws RemoteException {
	handle.switchReplicationStrategy(url, port_nb, strategy);
    }


    /**
     * Get the current replication policy of the remote task
     *
     * @return  the current replication policy of the remote task
     * @see ReplicationPolicy
     */
    public ReplicationPolicy getReplicationPolicy()
	throws RemoteException {
	return handle.getReplicationPolicy();
    }


    /**
     * Launches the election of a new leader.
     * This method is called internally when the current leader fails to answer.
     * <I>NB: in the absence of a proper FD, a leader failing to answer is
     * considered as having crashed</I>
     */
    private void findNewLeader() 
	throws RemoteException {
	System.out.println("Requesting new leader for task "+ task_name);
	try {
	    //Obtain access to naming services
	    Registry nserver_reg =
		LocateRegistry.getRegistry(NameServerImpl.URL, 
					   NameServerImpl.PORT_NB);
	    String nserver_name = 
		NameServerImpl.URL + ":" + NameServerImpl.PORT_NB 
		+ "/" + NameServer.SERVICE_NAME;
	    NameServer nserver = (NameServer) nserver_reg.lookup(nserver_name);
	    //Request new leader election
	    leader_info = nserver.selectAnotherLeader(leader_info);
	    if (leader_info == null)
		throw new DarxException();
	    System.out.println("NS elected " + leader_info.textifyDarxPath());
	    //Retrieve the reference to the new leader & set it as leader
	    Registry host_reg =
		LocateRegistry.getRegistry(leader_info.getURL(), 
					   leader_info.getPortNb());
	    TaskShellHandle new_leader_handle =
		(TaskShellHandle)host_reg.lookup(leader_info.textifyDarxName());
	    new_leader_handle.setAsGroupLeader();
	    handle = (DarxHandle) new_leader_handle;
	} catch (DarxException e) {
	    System.out.println("New leader for task "+ task_name 
			       + " cannot be found");
	    e.printStackTrace();
	} catch (NotBoundException e) {
	    System.out.println("RMI registration pb while fetching new leader for "+
			       task_name);
	    e.printStackTrace();	
	}    
    }
	    

    /**
     * Sets the replica at the given location as the new leader for its group.
     * NB: This includes triggering the notification of the new group
     * hierarchy to the concerned members (the old and the new leader)
     *
     * @param url      the url of the server hosting the new leader
     * @param port_nb  the port nb of the server hosting the new leader
     */
    public void setLeaderAt(String url, int port_nb)
	throws RemoteException {
	try {
	    //Obtain access to naming services
	    Registry nserver_reg =
		LocateRegistry.getRegistry(NameServerImpl.URL, 
					   NameServerImpl.PORT_NB);
	    String nserver_name = NameServerImpl.URL + ":" 
		+ NameServerImpl.PORT_NB + "/" + NameServer.SERVICE_NAME;
	    NameServer nserver = (NameServer) nserver_reg.lookup(nserver_name);
	    //Retrieve info for the old leader
	    ReplicantInfo old_leader_info = nserver.getLeaderOfTask(task_name);
	    //Elect the new leader through its location
	    ReplicantInfo new_leader_info =
		nserver.setLeaderOfTask(url, port_nb, task_name);
	    //Retrieve the interface for the old leader
	    Registry host_reg =
		LocateRegistry.getRegistry(old_leader_info.getURL(),
					   old_leader_info.getPortNb());
	    TaskShellHandle old_handle = 
		(TaskShellHandle) host_reg.lookup(old_leader_info.textifyDarxName());
	    //Retrieve the interface for the new leader
	    host_reg = LocateRegistry.getRegistry(url, port_nb);
	    TaskShellHandle new_handle = 
		(TaskShellHandle) host_reg.lookup(new_leader_info.textifyDarxName());
	    //Swap roles and notify the concerned replicas
	    //(namely the old and the new leader)
	    new_handle.setAsGroupLeader();
	    old_handle.demoteFromGroupLeader();
	    handle = (DarxHandle) new_handle;
	} catch (DarxException e) {
	    System.out.println("New leader for task "+ task_name 
			       + " cannot be elected");
	    e.printStackTrace();
	} catch (NotBoundException e) {
	    System.out.println("RMI registration pb while setting new leader for "+
			       task_name);
	    e.printStackTrace();	
	}   
    }

}
