/****
 *	Copyright (c) 2007, Olivier Marin, Laboratoire d'Informatique de Paris 6
 *	All rights reserved.
 ****/

/* 
	This file is part of DARX.

	DARX is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	DARX is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with DARX. If not, see <http://www.gnu.org/licenses/>.
 */




 

package darx;

import java.util.StringTokenizer;
import java.rmi.*;
import java.rmi.registry.*;
import java.io.*;
import java.net.MalformedURLException;


/**
 * This object contains the information regarding
 * a specific replicant handled by DARX.
 * It is used as a global name for each DarxTask.
 * 
 * 
 * @author Jacob Zimmermann
 * @author Olivier Marin
 * 
 * @version %I%, %G%
 **/
public class ReplicantInfo implements Serializable, Comparable {


    /**
     * the URL of the current host for this replica
     **/
    private String url;

    /**
     * the port number of the current host for this replica
     **/
    private int port_nb;

    /**
     * the generic task name to which this replica is related
     **/
    private String task_name;

    /**
     * the identifier for this replica
     **/
    private int id;


    /**
     * constructs the information container of a task's leader,
     * leaving the fields empty.
     **/
    ReplicantInfo() {
	this.url = "";
	this.port_nb = -1;
	this.task_name = "";
	this.id = -1;
    }


    /**
     * constructs the global identifier of a replica
     **/
    ReplicantInfo(String url, int port_nb, String task_name, int id) {
	this.url = url;
	this.port_nb = port_nb;
	this.task_name = task_name;
	this.id = id;
    }


    //SELECTORS

    /** @return the URL of the current host for this replica */
    public String getURL() { return url; }
    /** @return the port number of the current host for this replica */
    public int getPortNb() { return port_nb; }
    /** @return the generic task name to which this replica is related */
    public String getTaskName() { return task_name; }
    /** @return the identifier for this replica */
    public int getReplicantID() { return id; }


    //MODIFIERS

    public void setURL(String new_url) { url = new_url ; }
    public void setPortNb(int new_port_nb) { port_nb = new_port_nb ; }
    public void setTaskName(String new_task_name) { task_name = new_task_name ; }
    public void setReplicantID(int new_id) { id = new_id ; }



    //FORMAT MANIPULATIONS

    /**
     * parses the given DARX path name String
     * (format: <URL>:<port nb>/<generic task name>$<replication nb>)
     * in order to fill the fields of the current ReplicantInfo.
     */
    void parse(String info) {
	String temp;
	StringTokenizer st1 = new StringTokenizer(info, ":");
	url = st1.nextToken();
	temp = st1.nextToken();
	StringTokenizer st2 = new StringTokenizer(temp, "/");
	port_nb = (new Integer(st2.nextToken())).intValue();
	temp = st2.nextToken();
	StringTokenizer st3 = new StringTokenizer(temp, "$");
	task_name = st3.nextToken();
	id = (new Integer(st3.nextToken())).intValue();
    }


    /**
     * creates the unique DARX replicant name
     * corresponding to the current ReplicantInfo.
     */
    String textifyDarxName() {
	String temp = "";
	temp += task_name + "$" + id;
	return temp;
    }


    /**
     * creates the unique DARX path corresponding
     * to the current ReplicantInfo.
     */
    String textifyDarxPath() {
	String temp = "";
	temp += url + ":" + port_nb + "/"+ task_name+ "$" + id;
	return temp;
    }


    //------------------------//
    //------ COMPARISON ------//
    //------------------------//

    /**
     * Compares this replicant info to an object.
     * This method must be defined in order to implement interface Comparable.
     *
     * @param o  the object (info) to be compared
     * @return 0 if both infos correspond to the same replicant.
     */
    public int compareTo(Object o) {
	ReplicantInfo ri = (ReplicantInfo)o;
	return (this.textifyDarxPath().compareTo(ri.textifyDarxPath()));
    }


    /**
     * Compares the specified Object with this info for equality.
     *
     * @param o  the object to be compared (hopefully a 
     *           <code>ReplicationInfo</code>)
     * @return true if the specified Object is equal to this info.
     */
    public boolean equals(Object o) {
	return (compareTo((ReplicantInfo)o) == 0);
    }


    /**
     * Who knows when THIS might come in handy...
     */
    public int hashCode() {
	return (this.textifyDarxPath()).hashCode();
    }



    /**
     * Returns whether the given location is that of the current ReplicantInfo.
     */
    public boolean isAt(String l_url, int l_port_nb) {
	return ((url.compareTo(l_url) ==0) && (port_nb == l_port_nb));
    }


    /**
     * Retrieves the remote reference to the given replica, ie. the
     * <code>TaskShellHandle</code> corresponding to the
     * replicant information.
     */
    TaskShellHandle getTaskShellHandle() throws RemoteException {
	TaskShellHandle tsh = null;
	try {
	    Registry host_reg =
		LocateRegistry.getRegistry(url, port_nb);
	    tsh = (TaskShellHandle) host_reg.lookup(textifyDarxName());
	} catch (DarxException e) {
	    System.out.println("Shell handle for replicant "+ textifyDarxPath()
			       + " cannot be found");
	    e.printStackTrace();
	} catch (NotBoundException e) {
	    System.out.println("RMI registration pb while fetching handle for " +
			       textifyDarxPath());
	    e.printStackTrace();	
	}
	return tsh;
    }

}

