/****
 *	Copyright (c) 2007, Olivier Marin, Laboratoire d'Informatique de Paris 6
 *	All rights reserved.
 ****/

/* 
	This file is part of DARX.

	DARX is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	DARX is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with DARX. If not, see <http://www.gnu.org/licenses/>.
 */




 


package darx;


import java.util.*;
import java.io.Serializable;

/**
 * This object contains the information concerning
 * a specific replication group handled by DARX.
 * It provides the leader of the replication group, 
 * as well as the list of all the members.
 * 
 * 
 * @author Olivier Marin
 * 
 * @version %I%, %G%
 **/
public class ReplicationGroupInfo implements Serializable {



    /**
     * the name of the application task represented by the replication group.
     */
    private String task_name;

    /**
     * the identification of the replication group leader.
     * A value of -1 means that the group is empty.
     */
    private int leader;

    /**
     * the list of the replication group members.
     * key: <int> replicant number
     * accessed info: <ReplicantInfo> replicant global naming info
     */
    private Hashtable members;



    /**
     * Constructs an empty information container.
     */
    public ReplicationGroupInfo() {
	task_name = "";
	leader = -1;
	members = new Hashtable();
    }


    /**
     * Constructs an information container.
     */
    public ReplicationGroupInfo(String task_id, int leader_id, Hashtable member_list) {
	task_name = task_id;
	leader = leader_id;
	members = member_list;
    }


    /**
     * @return the name of the task represented by the group
     */
    public String getTaskName() { return task_name; }


    /**
     * Checks whether the given replicant info belongs to the group.
     *
     * @return true if it does belong to the group, false otherwise
     */
    public boolean contains(ReplicantInfo rep_info) {
	return members.contains(rep_info);
    }


    /**
     * @return the info concerning the current leader of the group
     */
    public ReplicantInfo getLeaderInfo() 
	throws NoMoreReplicantsException { 
	if (leader == -1)
	    throw new NoMoreReplicantsException(task_name);
	else
	    return (ReplicantInfo) members.get(new Integer(leader));
    }


    /**
     * Returns information about a replica at a specific location.
     *
     * @return the info concerning a given replica within the group
     */
    public ReplicantInfo getLocatedMemberInfo(String url, int port_nb)
	throws UnknownReplicantException {
	//Retrieve the list of info containers
	Enumeration e = members.elements();
	//Search through the list
	boolean found = false;
	ReplicantInfo info = null;
	while (!(found) && (e.hasMoreElements())) {
	    info = (ReplicantInfo) e.nextElement();
	    if ((url.compareTo(info.getURL()) == 0) && (info.getPortNb() == port_nb))
		found = true;
	}
	//If no corresponding replica was found
	if (found == false) {
	    info = new ReplicantInfo(url, port_nb, task_name, -1);
	    throw new UnknownReplicantException(info);
	} else
	    return info;
    }
	    

    public Enumeration getMembers() { return members.elements(); }


    public int getSize() { return members.size(); }
	

    //MODIFIERS

    public void setTaskName(String new_task_name) { 
	task_name = new_task_name; 
    }

    public void setLeader(int new_leader_id) { 
	leader = new_leader_id;
    }

    public void setLeader(ReplicantInfo new_leader) 
	throws UnknownReplicantException {
	if (this.contains(new_leader))
	    leader = new_leader.getReplicantID();
	else
	    throw new UnknownReplicantException(new_leader);
    }


    //GROUP HANDLING

    public void addMember(ReplicantInfo new_member) {
	Integer id = new Integer(new_member.getReplicantID());
	if (members.containsKey(id))
	    System.out.println(new_member.textifyDarxName()
		+" already registered in rep group");
	else {
	    members.put(id, new_member);
	    System.out.println(new_member.textifyDarxName()
		+" registered in rep group");
	}
    }

    public void addLeader(ReplicantInfo new_leader) {
	addMember(new_leader);
	leader = new_leader.getReplicantID();
    }


    /**
     * Removes a member from the replication group information.
     * Information on the leader cannot be removed unless it is the
     * last replica in the group.
     *
     * @param doomed_member the info regarding the removed replica
     */
    public void removeMember(ReplicantInfo doomed_member) 
	throws IllegalLeaderRemovalException, NoMoreReplicantsException {
	Integer doomed_id = new Integer(doomed_member.getReplicantID());
	if ((leader == doomed_id.intValue()) && (members.size() != 1))
	    throw new IllegalLeaderRemovalException(doomed_member);
	else if (leader == -1)
	    throw new NoMoreReplicantsException(task_name);
	else {
	    members.remove(doomed_id);
	    if (members.size() == 0)
		leader = -1;
	}
    }
	

    /**
     * Selects a new leader within the current replication group.
     * BEWARE: calling this method implies that the current leader
     * is considered faulty; the previous leader will automatically
     * be removed from the replication group.
     */
    public ReplicantInfo selectNewLeader()
	throws NoMoreReplicantsException {
	//If the group is empty or bound to be so
	if ((leader == -1) || (members.size() == 1)) {
	    //If the faulty replica is the last of its group
	    if (leader != -1)
		members.remove(new Integer(leader));
	    throw new NoMoreReplicantsException(task_name);
	}
	//Remove faulty leader
	members.remove(new Integer(leader));
	System.out.println("RGI: After leader removal, group size = "
			   + members.size());
	Integer new_leader_id = (Integer) members.keys().nextElement();
	System.out.println("RGI: old leader id = " + leader);
	leader = new_leader_id.intValue();
	System.out.println("RGI: new leader id = " + leader);
	//!!! Must warn new leader about its new role
	return (ReplicantInfo) members.get(new_leader_id);
    }


    /**
     * Displays the DARX path of every replicant in the group.
     */
    public void displayReplicants() {
	ReplicantInfo rep_info;
	for (Enumeration e = members.elements() ; e.hasMoreElements() ;) {
	    rep_info = (ReplicantInfo) e.nextElement();
	    System.out.println("\t" + rep_info.textifyDarxPath());
	}
    }

}
