/* 
This example presents the (un)replication feature as well as the replication switching feature.
A Sender agent emits messages to a Displayer agent via a Forwarder agent.
There are 4 phases:
phase 1: no replication
phase 2: the Displayer is replicated actively
phase 3: the strategy is switched back to passive
phase 4: the strategy is switched back to active

To execute this example, you need:
- 1 name server for which the address & port number are explicitly modified in "NameServer.java"
- 2 Darx servers launched with respect to the addresses used in the code below

For instance, the commands entered in our demo are: 
       <startnserv>
       <startdarx -ns plume.rsr.lip6.fr 7777>
       <startdarx -ns plume.rsr.lip6.fr 7777 -p 7000>
       <startdarx -ns plume.rsr.lip6.fr 7777 -p 8000>
*/ 


package testdarx;

import darx.*;
import java.io.*;
import java.rmi.*;
import java.rmi.server.*;



/*#############################
The Message Displayer receives asynchronous messages & displays them,
as well as the current counter value for the number of receptions
#############################*/
class MsgDisplayer extends DarxTask {
    private int count;
    public MsgDisplayer() { 
		super("MsgDisplayer"); 
		count=0;
	    System.out.println("Starting Displayer");
    }
    public void receiveAsyncMessage(Object msg) {
		String str=(String) msg;
		System.out.println("Message #"+(++count)+" received : "+str);
    }
}

/*#############################
Forwards the messages to the Message Displayer.
#############################*/
class MsgForward extends DarxTask {
    private DarxCommInterface comm;
    private RemoteTask displayer;

    public MsgForward() {
	super("MsgForward");
	comm=new DarxCommInterface(getTaskName());
	displayer=null;
    }
    public void receiveAsyncMessage(Object m) {
	if(displayer==null) {
	    try {
		displayer=this.findTask("MsgDisplayer");
	    }
	    catch(DarxException e) {
		System.out.println("Error : "+e);
	    }
	}
	System.out.println("forwarding message : "+(String) m);
	comm.sendAsyncMessage(displayer, (Serializable) m);
    }
}


/*#############################
Emits messages entered in the shell window
Its execution consists in 4 phases:
1. emission for a local MsgDisplayer 
2. emission for the same MsgDisplayer, replicated actively on another location
3. idem, but with a passive strategy
4. idem, but back with the active strategy
#############################*/
class MsgSender extends DarxTask implements Runnable {
    private DarxCommInterface comm;
    private RemoteTask forward, displayer;
    private Thread exec;

    public MsgSender(RemoteTask forward, RemoteTask displayer) {
	super("MsgSender");
	comm=new DarxCommInterface(getTaskName());
	this.forward=forward;
	this.displayer=displayer;
    }

    public void start() {
	System.out.println("Starting "+getTaskName());
	exec = new Thread(this);
	exec.start();
    }

    private void readUserInputs(int nb_of_inputs) {
	int k;
	BufferedReader console = 
	    new BufferedReader(new InputStreamReader(System.in));
	for(k=0 ; k < nb_of_inputs ; k++) {
	    System.out.println("Enter message to send :");
	    String msg = "";
	    try {
		msg = console.readLine();
	    } catch(IOException e) {
		System.out.println("IOEXception : "+e);
	    }
	    comm.sendAsyncMessage(forward, msg);
	}
    }

    public void run() {
	int nb_of_inputs = 1;
	readUserInputs(nb_of_inputs);
	try {
	    System.out.println("replicating displayer");
	    displayer.replicateTo(test.BACKUP_URL, test.BACKUP_PORT_NB);
	} catch(RemoteException e) {
	    System.out.println("Error during replication : "+e);
	}
	readUserInputs(nb_of_inputs);
	System.out.println("Switching to passive strategy");
	try {
	    displayer.switchReplicationStrategy(test.BACKUP_URL, test.BACKUP_PORT_NB, new PassiveReplicationStrategy(2000));
	} catch(RemoteException e) {}
	readUserInputs(nb_of_inputs);
	System.out.println("Switching back to active strategy");
	try {
	    displayer.switchReplicationStrategy(test.BACKUP_URL, test.BACKUP_PORT_NB, new ActiveReplicationStrategy());
	} catch(RemoteException e) {}
	readUserInputs(nb_of_inputs);
	try {
	    exec.sleep(2000);
	    displayer.killReplicantAt(test.BACKUP_URL, test.BACKUP_PORT_NB);
	} catch (Exception e) {
	    System.out.println("Couldn't finish properly: "+e);
	}
	System.exit(0);
    }
}

public class test {

    static public final String SENDER_URL = "plume.rsr.lip6.fr";
    static public final int SENDER_PORT_NB = 6789;

    static public final String FORWARDER_URL = "plume.rsr.lip6.fr";
    static public final int FORWARDER_PORT_NB = 6789;

    static public final String DISPLAYER_URL = "plume.rsr.lip6.fr";
    static public final int DISPLAYER_PORT_NB = 7000;

    static public final String BACKUP_URL = "plume.rsr.lip6.fr";
    static public final int BACKUP_PORT_NB = 8000;


    static public void main(String[] p) {
	MsgDisplayer displayer = new MsgDisplayer();
	MsgForward forward = new MsgForward();
	RemoteTask remote_displayer, remote_forward;
	try {
	    System.out.println("Starting Displayer");
	    remote_displayer = displayer.activateTask(DISPLAYER_URL, 
						DISPLAYER_PORT_NB);
	    System.out.println("Starting Forwarder");
	    remote_forward = forward.activateTask(FORWARDER_URL, 
						  FORWARDER_PORT_NB);
	    MsgSender sender = new MsgSender(remote_forward, remote_displayer);
	    System.out.println("Starting Sender");
	    sender.activateTask(WRITER_URL, WRITER_PORT_NB);
	}
	catch(Exception e) {
	    System.out.println("Startup error : "+e);
	    //e.printStackTrace();
	}
    }
}
